# -*- coding: utf-8 -*-

'''
Run worker functions in separate processes.

The use case is somewhat similar to what can be done using either :class:`queue.Queue <Queue.Queue>` and threads, or using :mod:`multiprocessing`. That is: run a number of independent job functions, dispatched over a number of worker threads/processes.

The mpfork module is useful in cases the above methods cannot work:

* threading in python is globally mainly inefficient because of the GIL.
* :mod:`multiprocessing` makes heavy use of pickles to pass objects and parameters, and there are cases we are using objects which cannot be pickled, or which pickling generates heavy IO traffic (large data objects)

The method here is based on the queue / thread schema, but uses fork() to actually execute the workers in a separate process. Only results are passed through pickles, so the worker return results must be picklable, but not the input arguments and objects.

Use:

* allocate a :class:`Queue <Queue.Queue>`
* allocate a results list with the exact size of the jobs number
* allocate a set of worker threads (typically one per processor or core). The function :func:`allocate_workers` can do this for you.
* fill the queue with jobs, each being a tuple (job_index, function, args, kwargs, results_list)
* add a number of empty jobs (None) to the queue, one per allocated worker: this will be the marker for the end of processing in each worker thread. It is necessary to explicitly add these empty jobs since an empty queue is not the signal for the end of processing: it can be re-filled at any time.
* jobs run in workers
* join the queue
* join the worker threads
* the result list gets the return data for each job

::

    njobs = 10
    q = queue.Queue()
    res = [None] * njobs
    workers = allocate_workers(q, 0) # one per core
    for i in range(njobs):
        job = (i, sum, ((i, i), ), {}, res)
        q.put(job)

    # add as many empty jobs as the workers number to end them
    for i in range(len(workers)):
        q.put(None)

    # wait for every job to complete
    q.join()
    # terminate all threads
    for w in workers:
        w.join()

    print('result:', res)

In case of error, the job result will be an exception with stack information: (exception type, exception instance, stack_info)


Availability: Unix
'''

from __future__ import print_function
from __future__ import absolute_import
import multiprocessing
import threading
import queue
import os
import tempfile
import sys
from six.moves import range
try:
    import cpickle as pickle
except ImportError:
    import pickle

def run_job(f, *args, **kwargs):
    ''' Internal function, runs the function in a remote process.
    Uses fork() to perform it.

    Availability: Unix
    '''
    out_file = tempfile.mkstemp()
    os.close(out_file[0])
    pid = os.fork()
    if pid != 0:
        # parent: wait for the child
        pid, status = os.waitpid(pid, 0)
        # read output file
        #print('read from', os.getpid(), ':', out_file[1])
        if os.stat(out_file[1]).st_size == 0:
            # child did not write anything
            os.unlink(out_file[1])
            raise OSError('child did not output anything')
        if status != 0:
            os.unlink(out_file[1])
            raise RuntimeError('subprocess error: %d' % status)
        result = pickle.load(open(out_file[1], 'rb'))
        os.unlink(out_file[1])
        # traceback objects cannot be pickled...
        #if isinstance(result, tuple) and len(result) == 3 \
                #and isinstance(result[1], Exception):
            ## result is an axception with call stack: reraise it
            #raise result[0], result[1], result[2]
        if isinstance(result, Exception):
            raise result
        return result

    # child process
    try:
        try:
            #print('exec in', os.getpid(), ':', f, args, kwargs)
            result = f(*args, **kwargs)
            #print('OK')
        except Exception as e:
            # traceback objects cannot be pickled...
            #result = (type(e), e, sys.exc_info()[2])
            result = e
        #print('write:', out_file[1], ':', result)
        try:
            pickle.dump(result, open(out_file[1], 'wb'), protocol=2)
        except Exception as e:
            print('pickle failed:', e, '\nfor object:', type(result))
    finally:
        # sys.exit() is not enough
        os._exit(0)


def worker(q, thread_only, *args, **kwargs):
    ''' Internal function: worker thread loop:
    * pick a job in the queue q
    * execute it, either in the current thread, or in a remote process (using
      run_job()), depending on the thread_only parameter value
    * store the result in the result list
    * start again with another job

    The loop ends when a job in the queue is None.

    .. warning::
        Here we are making use of fork() (Unix only) inside a thread. Some systems do not behave well in this situation.
        See :func:`the os.fork() doc <os.fork>`
    '''
    while True:
        item = q.get()
        if item is None:
            q.task_done()
            break
        try:
            #print('run item in', os.getpid(), ':', item[:-1])
            sys.stdout.flush()
            i, f, argsi, kwargsi, res = item
            argsi = argsi + args
            kwargs = dict(kwargs)
            kwargs.update(kwargsi)
            try:
                if thread_only:
                    try:
                        result = f(*argsi, **kwargs)
                    except Exception as e:
                        result = e
                else:
                    result = run_job(f, *argsi, **kwargs)
                #print('result:', i, result)
                res[i] = result
            except Exception as e:
                res[i] = (type(e), e, sys.exc_info()[2])
                print(e)
        finally:
            #print('task done in', os.getpid(), ':', i, f)
            q.task_done()
            sys.stdout.flush()


def allocate_workers(q, nworker=0, thread_only=False, *args, **kwargs):
    ''' Utility finction to allocate worker threads.

    Parameters
    ----------
    q: :class:`Queue <Queue.Queue>` instance
        the jobs queue which will be fed with jobs for processing
    thread_only: bool
        if True, workers will run jobs in the worker thread, not using a
        separate process. This flag thus allows to choose a threaded or
        multiprocessing implementation.
    nworker: int
        number of worker threads (jobs which will run in parallel). A positive
        number (1, 2...) will be used as is, 0 means all available CPU cores
        (see :func:`multiprocessing.cpu_count`), and a negative number means
        all CPU cores except this given number.
    args, kwargs:
        additional arguments will be pased to the jobs function(s) after
        individual jobs arguments: they are args common to all jobs (if any)

    Returns
    -------
    workers: list
        workers list, each is a :class:`thread <threading.Thread>` instance
        running the worker loop function. Threads are already started (ie.
    '''
    if nworker == 0:
        nworker = multiprocessing.cpu_count()
    elif nworker < 0:
        nworker = multiprocessing.cpu_count() + nworker
        if nworker < 1:
            nworker = 1
    workers = []
    for i in range(nworker):
        w = threading.Thread(target=worker, args=(q, thread_only) + args,
                             kwargs=kwargs)
        w.start()
        workers.append(w)
    return workers



# -*- coding: utf-8 -*-

from __future__ import print_function

from __future__ import absolute_import
import unittest
import os
import sys
from soma import mpfork
import queue
import math
from six.moves import range


class TestMPFork(unittest.TestCase):

    if not sys.platform.startswith('win'):

        def test_mpfork(self):
            njobs = 10
            q = queue.Queue()
            res = [None] * njobs
            workers = mpfork.allocate_workers(q, 4)
            self.assertEqual(len(workers), 4)
            for i in range(njobs):
                job = (i, sum, ((i, i), ), {}, res)
                q.put(job)

            # add as many empty jobs as the workers number to end them
            for i in range(len(workers)):
                q.put(None)

            # wait for every job to complete
            q.join()
            # terminate all threads
            for w in workers:
                w.join()

            self.assertEqual(res, [i*2 for i in range(njobs)])

        def test_mpfork_with_exception1(self):
            njobs = 10
            q = queue.Queue()
            res = [None] * njobs
            workers = mpfork.allocate_workers(q, -10)
            self.assertTrue(len(workers) >= 1)
            for i in range(njobs):
                job = (i, math.sqrt, (i-2, ), {}, res)
                q.put(job)

            # add as many empty jobs as the workers number to end them
            for i in range(len(workers)):
                q.put(None)

            # wait for every job to complete
            q.join()
            # terminate all threads
            for w in workers:
                w.join()

            self.assertTrue(isinstance(res[0], tuple))
            self.assertTrue(res[0][0] is ValueError)
            self.assertTrue(isinstance(res[1], tuple))
            self.assertTrue(res[1][0] is ValueError)
            self.assertEqual(res[2:], [math.sqrt(i) for i in range(njobs - 2)])

        def test_mpfork_with_exception2(self):
            njobs = 10
            q = queue.Queue()
            res = [None] * njobs
            workers = mpfork.allocate_workers(q, 0)
            self.assertTrue(len(workers) >= 1)

            def job_func1(x):
                if x >= 2:
                    return x
                sys.exit(int(x))

            for i in range(njobs):
                job = (i, job_func1, (i, ), {}, res)
                q.put(job)

            # add as many empty jobs as the workers number to end them
            for i in range(len(workers)):
                q.put(None)

            # wait for every job to complete
            q.join()
            # terminate all threads
            for w in workers:
                w.join()

            #print(res, file=sys.stderr)
            self.assertTrue(isinstance(res[0], tuple))
            self.assertTrue(res[0][0] is OSError)
            self.assertTrue(isinstance(res[1], tuple))
            self.assertTrue(res[1][0] is OSError)
            self.assertEqual(res[2:], [i + 2 for i in range(njobs - 2)])


def test():
    suite = unittest.TestLoader().loadTestsFromTestCase(TestMPFork)
    runtime = unittest.TextTestRunner(verbosity=2).run(suite)
    return runtime.wasSuccessful()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

# System import
from __future__ import absolute_import
from __future__ import print_function
import re
import logging
import sys
import traits.api as traits
import six

# Define the logger
logger = logging.getLogger(__name__)

# Soma import
from soma.qt_gui.qt_backend import QtCore, QtGui
from .Str import StrControlWidget
import sip


class IntControlWidget(StrControlWidget):

    """ Control to enter an integer.
    """

    @staticmethod
    def is_valid(control_instance, *args, **kwargs):
        """ Method to check if the new control value is correct.

        If the new entered value is not correct, the backroung control color
        will be red.

        Parameters
        ----------
        control_instance: QLineEdit (mandatory)
            the control widget we want to validate

        Returns
        -------
        out: bool
            True if the control value is valid,
            False otherwise
        """
        # Get the current control palette
        control_palette = control_instance.palette()

        # Get the control current value: format the int string
        # Valid int strings are: +1, -1, 1
        control_text = control_instance.text()
        if type(control_text) not in (str, six.text_type):
            # old QString with PyQt API v1
            control_text = six.text_type(control_text)
        control_value = re.sub("^([-+])", "", control_text, count=1)

        red = QtGui.QColor(255, 220, 220)
        yellow = QtGui.QColor(255, 255, 200)

        # If the control value contains only digits, the control is valid and
        # the backgound color of the control is white
        is_valid = False
        if control_value.isdigit():
            control_palette.setColor(
                control_instance.backgroundRole(), QtCore.Qt.white)
            is_valid = True

        # If the control value is optional, the control is valid and the
        # backgound color of the control is yellow
        elif control_instance.optional is True and control_value == "":
            control_palette.setColor(control_instance.backgroundRole(), yellow)
            is_valid = True

        # If the control value is empty, the control is not valid and the
        # backgound color of the control is red
        else:
            control_palette.setColor(control_instance.backgroundRole(), red)

        # Set the new palette to the control instance
        control_instance.setPalette(control_palette)

        return is_valid

    @staticmethod
    def update_controller(controller_widget, control_name, control_instance,
                          reset_invalid_value=False, *args, **kwargs):
        """ Update one element of the controller.

        At the end the controller trait value with the name 'control_name'
        will match the controller widget user parameters defined in
        'control_instance'.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QLineEdit (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Update the controller only if the control is valid
        if IntControlWidget.is_valid(control_instance):

            # Get the control value
            if control_instance.text() == "":
                new_trait_value = traits.Undefined
            else:
                new_trait_value = int(control_instance.text())

            protected = controller_widget.controller.is_parameter_protected(
                control_name)
            # value is manually modified: protect it
            if getattr(controller_widget.controller, control_name) \
                    != new_trait_value:
                controller_widget.controller.protect_parameter(control_name)
            # Set the control value to the controller associated trait
            try:
                setattr(controller_widget.controller, control_name,
                        new_trait_value)
                logger.debug(
                    "'IntControlWidget' associated controller trait '{0}' "
                    "has been updated with value '{1}'.".format(
                        control_name, new_trait_value))
                return
            except traits.TraitError as e:
                print(e, file=sys.stderr)
                if not protected:
                    controller_widget.controller.unprotect_parameter(
                        control_name)

        if reset_invalid_value:
            # invalid, reset GUI to older value
            old_trait_value = getattr(controller_widget.controller,
                                      control_name)
            if old_trait_value is traits.Undefined:
                control_instance.setText("")
            else:
                control_instance.setText(six.text_type(old_trait_value))

    @staticmethod
    def update_controller_widget(controller_widget, control_name,
                                 control_instance):
        """ Update one element of the controller widget.

        At the end the controller widget user editable parameter with the
        name 'control_name' will match the controller trait value with the same
        name.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QLineEdit (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """

        try:
            test = control_instance.setText
        except ReferenceError:
            # widget deleted in the meantime
            return

        if sip.isdeleted(control_instance.__init__.__self__):
            IntControlWidget.disconnect(controller_widget, control_name,
                                         control_instance)
            return

        # Get the trait value
        new_controller_value = getattr(
            controller_widget.controller, control_name, traits.Undefined)

        # Set the trait value to the int control
        if new_controller_value is traits.Undefined:
            control_instance.setText("")
        else:
            control_instance.setText(six.text_type(new_controller_value))
        logger.debug("'IntControlWidget' has been updated with value "
                     "'{0}'.".format(new_controller_value))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import
import os
from setuptools import setup, find_packages

python_dir = os.path.join(os.path.dirname(__file__), "python")
release_info = {}

with open(os.path.join(python_dir, "soma", "info.py")) as f:
    code = f.read()
    exec(code, release_info)

setup(
    name=release_info["NAME"],
    description=release_info["DESCRIPTION"],
    long_description=release_info["LONG_DESCRIPTION"],
    license=release_info["LICENSE"],
    classifiers=release_info["CLASSIFIERS"],
    author=release_info["AUTHOR"],
    author_email=release_info["AUTHOR_EMAIL"],
    version=release_info["VERSION"],
    package_dir = {'': python_dir},
    packages=find_packages(python_dir),
    platforms=release_info["PLATFORMS"],
    install_requires=release_info["REQUIRES"],
    extras_require = release_info["EXTRAS_REQUIRE"],
)



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

from __future__ import absolute_import
import string
import sys


def load_objects(module_name, object_name=None, allowed_instances=None):
    """ Load a python object

    The object is described by a string `module_name` and `object_name`.
    If `object_name` is None import the full module content.
    Can also filtered objects and only keep objects of
    types defined in allowed_instances.

    Parameters
    ----------
    module : str
        module name
    tool : str
        tool name
    allowed_instances : list
        list of instance to load

    Returns
    -------
    tools : list
        a list of objects.
    """
    # remove special characters
    module_name = cleanup(module_name)
    if object_name:
        object_name = cleanup(object_name)

    # import the module
    __import__(module_name)
    module = sys.modules[module_name]

    # get the target tool(s)
    tools = []
    if object_name:
        try:
            insert_tool(tools, getattr(module, object_name),
                        allowed_instances)
        except ImportError as e:
            raise Exception("Could not import {0}: {1}".format(object_name,
                                                               e))
    else:
        for tool_name in dir(module):
            if tool_name.startswith("_"):
                continue
            try:
                insert_tool(tools, getattr(module, tool_name),
                            allowed_instances)
            except ImportError as e:
                raise Exception("Could not import {0}: {1}".format(
                                tool_name, e))

    return tools


def cleanup(attribute):
    """ cleanup avoiding:

        * Windows reserved characters
        * '_' since it is reserved by Brainvisa
        * tab, newline and null character
    """
    cleanup_table = string.maketrans('(){} \t\r\n\0', '         ')

    attribute = attribute.translate(cleanup_table)
    attribute = attribute.replace(" ", "")

    return attribute


def insert_tool(tools, tool, allowed_instances):
    """ Add tool to list if tool is sub class
    of one item in allowed_instances
    """
    allowed_instances = allowed_instances or [object, ]
    for check_instance in allowed_instances:
        if isinstance(tool, type) and issubclass(tool, check_instance):
            tools.append(tool)
            break



# -*- coding: utf-8 -*-
from __future__ import print_function

from __future__ import absolute_import
from .load_pilots import load_pilots
import soma
import sys

error_message = """

##################################
#           GAME OVER            #
##################################

              +---+
              |   |
              O   |
             /|\  |
             / \  |
                  |
            =========
"""

valid_message = """

##################################
#        MODULE TEST PASSED      #
##################################

              +---+
              |   |
                  |
                  | \O/
                  |  |
                  | / \\
           ==============
            """


def run_all_tests():
    """ Execute all the unitests.

    Returns
    -------
    is_valid : bool
        True if all the tests are passed successfully,
        False otherwise.
    """
    module_path = soma.__path__[0]
    tests = load_pilots(module_path, module_path)

    is_valid = True
    for module, ltest in tests.items():
        for test in ltest:
            is_valid = is_valid and test()

    return is_valid


def is_valid_module():
    is_valid = run_all_tests()
    if is_valid:
        print(valid_message)
        return True
    else:
        print(error_message)
        return False


if __name__ == "__main__":
    res = is_valid_module()
    if not res:
        sys.exit(1)



# -*- coding: utf-8 -*-
#
# Soma-base - Copyright (C) CEA, 2013
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

'''Compatibility module for PyQt and PySide. Currently supports PyQt4,
PySide, and PyQt5.
This modules handles differences between PyQt and PySide APIs and behaviours,
and offers a few functions to make it easier to build neutral GUI code, which
can run using either backend.

The main funcion here is set_qt_backend() which must be called to initialize
the appropriate backend. Most functions of this module assume set_qt_backend()
has been called first to setup internal variables.

Note that such compatibility generally requires to use PyQt4 with SIP API
version 2, ie do not use QString, QVariant, QDate and similar classes, but
directly convert to/from python types, which is also PySide behaviour. The
qt_backend module switches to this API level 2, but this only works before the
PyQt modules are imported, thus it may fail if PyQt has already be imported
without such settings.

Qt submodules can be imported in two ways:

>>> from soma.qt_gui import qt_backend
>>> qt_backend.import_qt_submodule('QtWebKit')

or using the import statement:

>>> from soma.qt_gui.qt_backend import QtWebKit

in the latter case, set_qt_backend() will be called automatically to setup the
appropriate Qt backend, so that the use of the backend selection is more
transparent.
'''

from __future__ import absolute_import
import logging
import sys
import os
import imp
import importlib
import inspect
import six

getfullargspec = getattr(inspect, 'getfullargspec',
                         getattr(inspect, 'getargspec', None))


# make qt_backend a fake module package, with Qt modules as sub-modules
__package__ = __name__
__path__ = [os.path.dirname(__file__)]

# internal variable to avoid warning several times
_sip_api_set = False

qt_backend = None
make_compatible_qt5 = False


class QtImporter(object):

    def find_module(self, fullname, path=None):
        modsplit = fullname.split('.')
        modpath = '.'.join(modsplit[:-1])
        module_name = modsplit[-1]
        if modpath != __name__ or module_name == 'sip':
            return None
        set_qt_backend()
        qt_backend = get_qt_backend()
        qt_module = get_qt_module()
        if make_compatible_qt5 and qt_backend in ('PyQt4', 'PySide'):
            if module_name == 'QtWidgets':
                module_name = 'QtGui'
            elif module_name == 'QtWebKitWidgets':
                module_name = 'QtWebKit'
        if qt_backend in ('PySide', 'PyQt6') and module_name == 'Qt':
            module_name = 'QtGui'
        found = imp.find_module(module_name, qt_module.__path__)
        return self

    def load_module(self, name):
        qt_backend = get_qt_backend()
        module_name = name.split('.')[-1]
        imp_module_name = module_name
        if make_compatible_qt5:
            if module_name == 'QtWidgets':
                imp_module_name = 'QtGui'
            elif module_name == 'QtWebKitWidgets':
                imp_module_name = 'QtWebKit'

        if qt_backend in('PyQt4', 'PyQt5', 'PyQt6'):
            # import the right sip module
            try:
                __import__('%s.sip' % qt_backend)
                sip = sys.modules['%s.sip' % qt_backend]
                sys.modules['sip'] = sip
            except:
                import sip

        if imp_module_name == 'Qt' and (qt_backend in ('PySide', 'PyQt6')
                                        or sip.SIP_VERSION >= 0x060000):
            # PySide and PyQt6 don't define the aggregating Qt module
            psmods = []
            base = name.split('.')[:-1]
            qt_mod = sys.modules[qt_backend]
            mods = set([x.split('.')[0]
                        for x in os.listdir(os.path.dirname(qt_mod.__file__))
                        if not x.startswith('_')])
            #for mod in ('QtCore', 'QtGui', 'phonon', 'QtNetwork', 'QtSvg',
                        #'QtOpenGL', 'QtTest', 'QtDeclarative', 'QtScript',
                        #'QtUiTools', 'QtScriptTools', 'QtWebKit', 'QtHelp',
                        #'QtSql', 'QtXml'):
            if 'Qt' in mods:
                # PyQt5 + sip6 brings a Qt module, but which is empty. We thus
                # need to re-populate it as in PyQt6, but here the module
                # already exists: don't reload it.
                mods.remove('Qt')
            for mod in mods:
                try:
                    psmods.append(self.load_module('.'.join(base + [mod])))
                except ImportError:
                    pass
            patch_main_modules(psmods)
            mod = sys.modules['.'.join([qt_backend, 'Qt'])]
            sys.modules['soma.qt_gui.qt_backend.Qt'] = mod
            return mod

        __import__('.'.join([qt_backend, imp_module_name]))
        module = sys.modules['.'.join([qt_backend, imp_module_name])]
        # fixes: #13432 - Ubuntu 14.04 LTS: Importing some modules
        #                                   of scikit learn from
        #                                   brainvisa process raises segfault
        # ref: https://bioproj.extra.cea.fr/redmine/issues/13432
        if module_name == 'uic' and qt_backend == 'PyQt4':
            def _safe_load_plugin(plugin, plugin_globals, plugin_locals):
                def _safe_getFilter():
                    import sys, DLFCN
                    res = plugin_locals['getFilter_orig']()
                    sys.setdlopenflags(DLFCN.RTLD_NOW)

                    return res

                import os
                __import__('.'.join(['PyQt4', 'uic', 'objcreator']))
                uic = sys.modules['.'.join([qt_backend, module_name])]
                res = uic.objcreator.load_plugin_orig(plugin,
                                                      plugin_globals,
                                                      plugin_locals)

                # It seems that this function is sometimes called with a first
                # argument of type File, sometimes of type str. Both cases
                # should be handled by this switch.
                if hasattr(plugin, "name"):
                    filename = plugin.name
                else:
                    filename = plugin
                if os.path.splitext(os.path.basename(filename))[0] == 'kde4':
                    # Replaces kde4 getFilter function
                    if ('getFilter_orig' not in plugin_locals):
                        plugin_locals['getFilter_orig'] \
                            = plugin_locals['getFilter']
                        plugin_locals['getFilter'] = _safe_getFilter

                return res

            __import__('.'.join([qt_backend, module_name, 'objcreator']))
            uic = sys.modules['.'.join([qt_backend, module_name])]
            # Replaces the load_plugin function in objcreator
            #uic.port_v2.load_plugin.load_plugin_orig \
                #= uic.port_v2.load_plugin.load_plugin
            #uic.port_v2.load_plugin.load_plugin = _safe_load_plugin
            if not hasattr(uic.objcreator, 'load_plugin_orig'):
                uic.objcreator.load_plugin_orig \
                    = uic.objcreator.load_plugin
                uic.objcreator.load_plugin = _safe_load_plugin

        sys.modules[name] = module
        if make_compatible_qt5:
            if imp_module_name == 'QtGui':
                from . import QtCore
                if qt_backend in ('PyQt4', 'PySide'):
                    sys.modules['.'.join([qt_backend, 'QtWidgets'])] = module
                    patch_qt4_modules(QtCore, module)
                elif qt_backend in('PyQt5', 'PyQt6'):
                    __import__('.'.join([qt_backend, 'QtWidgets']))
                    qtwidgets = sys.modules['.'.join([qt_backend,
                                                      'QtWidgets'])]
                    patch_qt5_modules(QtCore, module, qtwidgets)
                    if module_name == 'QtWidgets':
                        module = qtwidgets
            elif imp_module_name == 'QtWebKit':
                if qt_backend in ('PyQt4', 'PySide'):
                    sys.modules['.'.join([qt_backend, 'QtWebKitWidgets'])] \
                        = module
                elif qt_backend in('PyQt5', 'PyQt6'):
                    __import__('.'.join([qt_backend, 'QtWebKitWidgets']))
                    qtwebkitwidgets = sys.modules[
                        '.'.join([qt_backend,'QtWebKitWidgets'])]
                    patch_qt5_webkit_modules(module, qtwebkitwidgets)
                    if module_name == 'QtWebKitWidgets':
                        module = qtwebkitwidgets

        return module


# tune the import statement to get Qt submodules in this one
sys.meta_path.append(QtImporter())


def get_qt_backend():
    '''get currently setup or loaded Qt backend name: "PyQt4" or "PySide"'''
    global qt_backend
    if qt_backend is None:
        pyside = sys.modules.get('PySide')
        if pyside is not None:
            qt_backend = 'PySide'
        else:
            backends = ('PyQt6', 'PyQt5', 'PyQt4')
            for backend in backends:
                pyqt = sys.modules.get(backend)
                if pyqt is not None:
                    qt_backend = backend
                    break
    return qt_backend


def set_qt_backend(backend=None, pyqt_api=1, compatible_qt5=None):
    '''set the Qt backend.

    If a different backend has already setup or loaded, a warning is issued.
    If no backend is specified, try to guess which one is already loaded.

    If no backend is loaded yet, try to behave like IPython does.
    See: https://ipython.org/ipython-doc/dev/interactive/reference.html#pyqt-and-pyside

    More precisely this means:
    * If QT_API environement variable is not set, use PyQt4, with PyQt API v1
    * if QT_API is set to "pyqt" or "pyqt4", use PyQt4, with PyQt API v2
    * if QT_API is set to "pyside", use PySide
    * if QT_API is set to "pyqt5", use PyQt5
    * if QT_API is set to "pyqt6", use PyQt6

    Moreover if using PyQt4, QtCore is patched to duplicate QtCore.pyqtSignal
    and QtCore.pyqtSlot as QtCore.Signal and QtCore.Slot. This is meant to ease
    code portability between both worlds.

    if compatible_qt5 is set to True, modules QtGui and QtWidgets will be
    exposed and completed to contain the same content, with both Qt4 and Qt5.

    Parameters
    ----------
    backend: str (default: None)
        name of the backend to use
    pyqt_api: int (default: 1)
        PyQt API version: 1 or 2, only useful for PyQt4
    compatible_qt5: bool (default: None)
        expose QtGui and QtWidgets with the same content.
        If None (default), do not change the current setting.
        If True, in Qt5, when QtGui or QtWidgets is loaded, the other module
        (QtWidgets or QtGui) is also loaded, and the QtGui module is modified
        to contain also the contents of QtWidgets, so as to have more or less
        the same elements as in Qt4. It is a bit dirty and crappy but allows
        the same code to work with both versions of Qt.
        In Qt4, when QtGui is loaded, the module is also registered as
        QtWidgets, so QtGui and QtWidgets are the same module. Loading
        QtWidgets will also bring QtGui.

    Examples
    --------
        >>> from soma.qt_gui import qt_backend
        >>> qt_backend.set_qt_backend('PySide')
        >>> qt_backend.import_qt_submodule('QtCore')
        <module 'PySide.QtCore' from '/usr/lib/python2.7/dist-packages/PySide/QtCore.so'>
    '''
    global qt_backend
    global make_compatible_qt5
    qt5_compat_changed = False
    if compatible_qt5 is not None:
        make_compatible_qt5 = compatible_qt5
        qt5_compat_changed = True
    get_qt_backend()
    if backend is None:
        if qt_backend is None:
            # try to get from the environment variable QT_API, complying to
            # ETS 4
            # see
            # https://ipython.org/ipython-doc/dev/interactive/reference.html#pyqt-and-pyside
            qt_api = os.getenv('QT_API')
            if qt_api == 'pyqt6':
                backend = 'PyQt6'
            elif qt_api == 'pyqt5':
                backend = 'PyQt5'
            elif qt_api in ('pyqt', 'pyqt4'):
                backend = 'PyQt4'
                pyqt_api = 2
            elif qt_api == 'pyside':
                backend = 'PySide'
            else:
                backend = 'PyQt4'
                pyqt_api = 1
        else:
            backend = qt_backend
    if qt_backend is not None and qt_backend != backend:
        logging.warn('set_qt_backend: a different backend, %s, has already '
                     'be set, and %s is now requested' % (qt_backend, backend))
    if backend == 'PyQt4':  # and sys.modules.get('PyQt4') is None:
        import sip
        if pyqt_api == 2:
            sip_classes = ['QString', 'QVariant', 'QDate', 'QDateTime',
                           'QTextStream', 'QTime', 'QUrl']
            global _sip_api_set
            for sip_class in sip_classes:
                try:
                    sip.setapi(sip_class, pyqt_api)
                except ValueError as e:
                    if not _sip_api_set:
                        logging.warning(e.message)
            _sip_api_set = True
    qt_module = __import__(backend)
    __import__(backend + '.QtCore')
    #__import__(backend + '.QtGui')
    qt_backend = backend

    if qt_backend in('PyQt4', 'PyQt5', 'PyQt6'):
        # import the right sip module
        try:
            __import__('%s.sip' % qt_backend)
            sip = sys.modules['%s.sip' % qt_backend]
            sys.modules['sip'] = sip
        except:
            import sip

    if make_compatible_qt5 and qt5_compat_changed:
        ensure_compatible_qt5()
    else:
        if backend in('PyQt4', 'PyQt5', 'PyQt6'):
            qt_module.QtCore.Signal = qt_module.QtCore.pyqtSignal
            qt_module.QtCore.Slot = qt_module.QtCore.pyqtSlot


def patch_qt5_modules(QtCore, QtGui, QtWidgets):
    # copy QtWidgets contents into QtGui
    for key in QtWidgets.__dict__:
        if not key.startswith('__') and key not in QtGui.__dict__:
            setattr(QtGui, key, getattr(QtWidgets, key))
    # more hacks
    QtGui.QSortFilterProxyModel = QtCore.QSortFilterProxyModel
    QtGui.QItemSelectionModel = QtCore.QItemSelectionModel


def patch_qt5_webkit_modules(QtWebKit, QtWebKitWidgets):
    # copy QtWebKitWidgets contents into QtWebKit
    for key in QtWebKitWidgets.__dict__:
        if not key.startswith('__') and key not in QtWebKit.__dict__:
            setattr(QtWebKit, key, getattr(QtWebKitWidgets, key))


def patch_qt4_modules(QtCore, QtGui):
    QtCore.QSortFilterProxyModel = QtGui.QSortFilterProxyModel
    QtCore.QItemSelectionModel = QtGui.QItemSelectionModel


def patch_main_modules(modules):
    if '%s.Qt' % qt_backend in sys.modules:
        Qt = sys.modules['%s.Qt' % qt_backend]
    else:
        Qt = imp.new_module('%s.Qt' % qt_backend)
        sys.modules['%s.Qt' % qt_backend] = Qt
    for mod in modules:
        for key, item in six.iteritems(mod.__dict__):
            if not key.startswith('__') and key not in Qt.__dict__:
                setattr(Qt, key, item)


def ensure_compatible_qt5():
    if not make_compatible_qt5:
        return
    qt_backend = get_qt_backend()
    if qt_backend in ('PyQt5', 'PyQt6'):
        qtgui = None
        qtwidgets = None
        qtwebkit = None
        qtwebkitwidgets = None
        if '%s.QtGui' % qt_backend in sys.modules:
            qtgui = sys.modules['%s.QtGui' % qt_backend]
        if '%s.QtWidgets' % qt_backend in sys.modules:
            qtwidgets = sys.modules['%s.QtWidgets' % qt_backend]
        if '%s.QtWebKit' % qt_backend in sys.modules:
            qtwebkit = sys.modules['%s.QtWebKit' % qt_backend]
        if '%s.QtWebKitWidgets' % qt_backend in sys.modules:
            qtwebkitwidgets = sys.modules['%s.QtWebKitWidgets' % qt_backend]
        if qtgui and qtwidgets is None:
            from . import QtWidgets
            qtwidgets = sys.modules['%s.QtWidgets' % qt_backend]
        elif qtwidgets and qtgui is None:
            from . import QtGui
            qtgui = sys.modules['%s.QtGui' % qt_backend]
        elif qtgui and qtwidgets:
            from . import QtCore
            patch_qt5_modules(QtCore, qtgui, qtwidgets)
        if qtwebkit and qtwebkitwidgets is None:
            from . import QtWebKitWidgets
            qtwebkitwidgets = sys.modules['%s.QtWebKitWidgets' % qt_backend]
        elif qtwebkitwidgets and qtwebkit is None:
            from . import QtWebKit
        elif qtwebkit and qtwebkitwidgets:
            patch_qt5_webkit_modules(qtwebkit, qtwebkitwidgets)
    else:
        if '%s.QtGui' % qt_backend in sys.modules:
            from . import QtWidgets
        from . import QtCore, QtGui
        patch_qt4_modules(QtCore, QtGui)
    if qt_backend in('PyQt4', 'PyQt5', 'PyQt6'):
        from . import QtCore
        QtCore.Signal = QtCore.pyqtSignal
        QtCore.Slot = QtCore.pyqtSlot

    # export enums
    from soma.utils.sip_compat import sip_export_enums
    pyqt = sys.modules[qt_backend]
    sip_export_enums(pyqt)



def get_qt_module():
    '''Get the main Qt module (PyQt4 or PySide)'''
    global qt_backend
    return sys.modules.get(qt_backend)


def import_qt_submodule(submodule):
    '''Import a specified Qt submodule.
    An alternative to the standard statement:

    >>> from soma.qt_gui.qt_backend import <submodule>

    The main differences is that it forces loading the module from the
    appropriate backend, whereas the import statement will reuse the already
    loaded one. Moreover it returns the module.

    For instance,

    >>> from soma.qt_gui import qt_backend
    >>> qt_backend.set_qt_backend('PyQt4')
    >>> from soma.qt_gui.qt_backend import QtWebKit
    >>> QtWebKit
    <module 'PyQt4.QtWebKit' from '/usr/lib/python2.7/dist-packages/PyQt4/QtWebKit.so'>
    >>> qt_backend.set_qt_backend('PySide') # changing backend
    WARNING:root:set_qt_backend: a different backend, PyQt4, has already be set, and PySide is now requested
    >>> from soma.qt_gui.qt_backend import QtWebKit
    >>> QtWebKit
    <module 'PyQt4.QtWebKit' from '/usr/lib/python2.7/dist-packages/PyQt4/QtWebKit.so'>

    In the above example, we are still using the QtWebKit from PyQt4.
    Now:

    >>> QtWebKit = qt_backend.import_qt_submodule('QtWebKit')
    >>> QtWebKit
    <module 'PySide.QtWebKit' from '/usr/lib/python2.7/dist-packages/PySide/QtWebKit.so'>

    We are now actually using PySide.
    Note that it is generally a bad idea to mix both...

    Parameters
    ----------
        submodule: str (mandatory)
            submodule name, ex: QtWebKit

    Returns
    -------
        the loaded submodule
    '''
    __import__(qt_backend + '.' + submodule)
    mod = sys.modules[qt_backend + '.' + submodule]
    return mod


def _iconset(self, prop):
    from . import QtGui
    return QtGui.QIcon(os.path.join(self._basedirectory,
                                    prop.text).replace("\\", "\\\\"))


def _pixmap(self, prop):
    from . import QtGui
    return QtGui.QPixmap(os.path.join(self._basedirectory,
                                      prop.text).replace("\\", "\\\\"))


def loadUi(ui_file, *args, **kwargs):
    '''Load a ``.ui`` file and returns the widget instance.

    This function is a replacement of PyQt4.uic.loadUi. The only difference is
    that relative icon or pixmap file names that are stored in the ``*.ui``
    file are considered to be relative to the directory containing the ui file.
    With PyQt4.uic.loadUi, relative file names are considered relative to the
    current working directory therefore if this directory is not the one
    containing the ui file, icons cannot be loaded.
    '''
    from . import QtGui
    if get_qt_backend() in ('PyQt4', 'PyQt5', 'PyQt6'):
        # the problem is corrected in version > 4.7.2,
        from . import QtCore
        if QtCore.PYQT_VERSION > 0x040702:
            from . import uic
            return uic.loadUi(ui_file, *args, **kwargs)
        else:
            # needed import and def
            from .uic.Loader import loader
            if not hasattr(globals(), 'partial'):
                from soma.functiontools import partial

            def _iconset(self, prop):
                return QtGui.QIcon(os.path.join(self._basedirectory, prop.text).replace("\\", "\\\\"))

            def _pixmap(self, prop):
                return QtGui.QPixmap(os.path.join(self._basedirectory, prop.text).replace("\\", "\\\\"))
            uiLoader = loader.DynamicUILoader()
            uiLoader.wprops._basedirectory = os.path.dirname(
                os.path.abspath(ui_file))
            uiLoader.wprops._iconset = partial(_iconset, uiLoader.wprops)
            uiLoader.wprops._pixmap = partial(_pixmap, uiLoader.wprops)
            return uiLoader.loadUi(ui_file, *args, **kwargs)
    else:
        from PySide.QtUiTools import QUiLoader
        return QUiLoader().load(ui_file)  # , *args, **kwargs )


def loadUiType(uifile, from_imports=False):
    '''PyQt4 / PySide abstraction to uic.loadUiType.
    Not implemented for PySide, actually, because PySide does not have this
    feature.
    '''
    if get_qt_backend() == 'PyQt6':
        from PyQt6 import uic
        return uic.loadUiType(uifile, from_imports=from_imports)
    if get_qt_backend() == 'PyQt5':
        from PyQt5 import uic
        return uic.loadUiType(uifile, from_imports=from_imports)
    if get_qt_backend() == 'PyQt4':
        # the parameter from_imports doesn't exist in our version of PyQt
        from PyQt4 import uic
        return uic.loadUiType(uifile)
    else:
        raise NotImplementedError('loadUiType does not work with PySide')
        # ui = loadUi(uifile)
        # return ui.__class__, QtGui.QWidget # FIXME


def getOpenFileName(parent=None, caption='', directory='', filter='',
                    selectedFilter=None, options=0):
    '''PyQt4 / PySide compatible call to QFileDialog.getOpenFileName'''
    set_qt_backend(compatible_qt5=True)
    from . import QtGui
    if get_qt_backend() in('PyQt4', 'PyQt5', 'PyQt6'):
        kwargs = {}
        # kwargs are used because passing None or '' as selectedFilter
        # does not work, at least in PyQt 4.10
        # On the other side I don't know if this kwargs works with older
        # sip/PyQt versions.
        if selectedFilter:
            kwargs['selectedFilter'] = selectedFilter
        if options:
            kwargs['options'] = QtGui.QFileDialog.Option(options)
        filename = get_qt_module().QtGui.QFileDialog.getOpenFileName(
            parent, caption, directory, filter, **kwargs)
        if get_qt_backend() == 'PyQt4':
            return filename
        else:
            return filename[0] # PyQt5 returns (filaname, filter)
    else:
        return get_qt_module().QtGui.QFileDialog.getOpenFileName(
            parent, caption, directory, filter, selectedFilter,
            QtGui.QFileDialog.Option(options))[0]


def getSaveFileName(parent=None, caption='', directory='', filter='',
                    selectedFilter=None, options=0):
    '''PyQt4 / PySide compatible call to QFileDialog.getSaveFileName'''
    set_qt_backend(compatible_qt5=True)
    from . import QtGui
    if get_qt_backend() in ('PyQt4', 'PyQt5', 'PyQt6'):
        kwargs = {}
        # kwargs are used because passing None or '' as selectedFilter
        # does not work, at least in PyQt 4.10
        # On the other side I don't know if this kwargs works with older
        # sip/PyQt versions.
        if selectedFilter:
            kwargs['selectedFilter'] = selectedFilter
        if options:
            kwargs['options'] = QtGui.QFileDialog.Option(options)
        filename = get_qt_module().QtGui.QFileDialog.getSaveFileName(
            parent, caption, directory, filter, **kwargs)
        if get_qt_backend() == 'PyQt4':
            return filename
        else:
            return filename[0] # PyQt5 returns (filaname, filter)
    else:
        return get_qt_module().QtGui.QFileDialog.getSaveFileName(
            parent, caption, directory, filter, selectedFilter, options)[0]


def getExistingDirectory(parent=None, caption='', directory='', options=None):
    '''PyQt4 / PySide compatible call to QFileDialog.getExistingDirectory'''
    set_qt_backend(compatible_qt5=True)
    from . import QtGui
    if get_qt_backend() in ('PyQt4', 'PyQt5', 'PyQt6'):
        kwargs = {}
        if options is not None:
            kwargs['options'] = QtGui.QFileDialog.Option(options)
        return get_qt_module().QtGui.QFileDialog.getExistingDirectory(
            parent, caption, directory, **kwargs)
    else:
        if options is not None:
            return get_qt_module().QtGui.QFileDialog.getExistingDirectory(
                parent, caption, directory,
                QtGui.QFileDialog.Option(options))[0]
        else:
            return get_qt_module().QtGui.QFileDialog.getExistingDirectory(
                parent, caption, directory)[0]


def init_matplotlib_backend(force=True):
    '''Initialize Matplotlib to use Qt, and the appropriate Qt/Python binding
    (PySide or PyQt) according to the configured/loaded toolkit.
    Moreover, the appropriate FigureCanvas type is set in the current module,
    and returned by this function.

    Parameters
    ----------
    force: bool
        if False, if the backend is already initialized with a different value,
        then raise an exception. If True (the default), force the new
        backend in matplotlib. If matplotlib does not support the force
        parameter, then the backend will not be forced.
    '''
    try:
        import matplotlib
    except ImportError:
        # if matplotlib cannot be found, don't do anything.
        return

    mpl_ver = [int(x) for x in matplotlib.__version__.split('.')[:2]]
    qt_backend = get_qt_backend()
    if qt_backend == 'PyQt6':
        guiBackend = 'Qt5Agg'  # apparently not Qt6Agg
        mpl_backend_mod = 'matplotlib.backends.backend_qt5agg'
    elif qt_backend == 'PyQt5':
        guiBackend = 'Qt5Agg'
        mpl_backend_mod = 'matplotlib.backends.backend_qt5agg'
    else:
        guiBackend = 'Qt4Agg'
        mpl_backend_mod = 'matplotlib.backends.backend_qt4agg'
    if 'matplotlib.backends' not in sys.modules or force:
        if six.PY3:
            argspec =inspect.getfullargspec(matplotlib.use)
            if 'force' in argspec.args or 'force' in argspec.kwonlyargs:
                matplotlib.use(guiBackend, force=force)
            else:
                matplotlib.use(guiBackend)
        else:
            if 'force' in getfullargspec(matplotlib.use).args:
                matplotlib.use(guiBackend, force=force)
            else:
                matplotlib.use(guiBackend)
    elif matplotlib.get_backend() != guiBackend:
        raise RuntimeError(
            'Mismatch between Qt version and matplotlib backend: '
            'matplotlib uses ' + matplotlib.get_backend() + ' but '
            + guiBackend + ' is required.')
    if qt_backend == 'PySide':
        if 'backend.qt4' in list(matplotlib.rcParams.keys()):
            # some versions (>=1.1, <3) of matplotlib have this rcParams setting
            matplotlib.rcParams['backend.qt4'] = 'PySide'
    else:
        if qt_backend == 'PyQt6':
            rc_key = 'backend.qt6'
        elif qt_backend == 'PyQt5':
            rc_key = 'backend.qt5'
        else:
            rc_key = 'backend.qt4'
        if rc_key in list(matplotlib.rcParams.keys()):
            # some versions (>=1.1, <3) of matplotlib have this rcParams setting
            matplotlib.rcParams[rc_key] = qt_backend
    __import__(mpl_backend_mod)
    backend_mod = sys.modules[mpl_backend_mod]
    FigureCanvas = backend_mod.FigureCanvasQTAgg
    sys.modules[__name__].FigureCanvas = FigureCanvas
    return mpl_backend_mod


traits_ui_handler_initialized = False


def init_traitsui_handler():
    ''' Setup handler for traits notification in Qt GUI.
    This function needs to be called before using traits notification which
    trigger GUI modification from non-principal threads.

    **WARNING**: depending on the Qt bindings (PyQt or PySide), this function
    may instantiate a QApplication. It seems that when using PyQt4,
    QApplication is not instantiated, whereas when using PySide, it is.
    This means that after this function has been called, one must check if
    the application has been created before recreating it:

    ::

        app = QtGui.QApplication.instance()
        if not app:
            app = QtGui.QApplication(sys.argv)

    This behaviour is triggered somewhere in the traitsui.qt4.toolkit module,
    we cannot change it easily.
    '''
    global traits_ui_handler_initialized
    from . import QtCore, QtGui
    if traits_ui_handler_initialized:
        return # already done

    try:
        if get_qt_backend() in ('PyQt4', 'PySide'):
            from traitsui.qt4 import toolkit
        else:
            # if using Qt5 we must not import traitsui.qt4, which would cause
            # a crash. Then use the code taken from traitsui.qt4.toolkit
            # in a qt-independent manner
            raise ImportError('traitsui doesn\'t provide a PyQt5 backend')
    except Exception:
        # copy of the code from traitsui.qt4.toolkit

        from traits.trait_notifiers import set_ui_handler

        #-------------------------------------------------------------------------------
        #  Handles UI notification handler requests that occur on a thread other than
        #  the UI thread:
        #-------------------------------------------------------------------------------
        _QT_TRAITS_EVENT = QtCore.QEvent.Type(QtCore.QEvent.registerEventType())

        class _CallAfter(QtCore.QObject):
            """ This class dispatches a handler so that it executes in the main GUI
                thread (similar to the wx function).
            """

            # The list of pending calls.
            _calls = []

            # The mutex around the list of pending calls.
            _calls_mutex = QtCore.QMutex()

            def __init__(self, handler, *args, **kwds):
                """ Initialise the call.
                """
                QtCore.QObject.__init__(self)

                # Save the details of the call.
                self._handler = handler
                self._args = args
                self._kwds = kwds

                # Add this to the list.
                self._calls_mutex.lock()
                self._calls.append(self)
                self._calls_mutex.unlock()

                # Move to the main GUI thread.
                self.moveToThread(QtGui.QApplication.instance().thread())

                # Post an event to be dispatched on the main GUI thread. Note that
                # we do not call QTimer.singleShot, which would be simpler, because
                # that only works on QThreads. We want regular Python threads to work.
                event = QtCore.QEvent(_QT_TRAITS_EVENT)
                QtGui.QApplication.instance().postEvent(self, event)

            def event(self, event):
                """ QObject event handler.
                """
                if event.type() == _QT_TRAITS_EVENT:
                    # Invoke the handler
                    self._handler(*self._args, **self._kwds)

                    # We cannot remove from self._calls here. QObjects don't like being
                    # garbage collected during event handlers (there are tracebacks,
                    # plus maybe a memory leak, I think).
                    QtCore.QTimer.singleShot(0, self._finished)

                    return True
                else:
                    return QtCore.QObject.event(self, event)

            def _finished(self):
                """ Remove the call from the list, so it can be garbage collected.
                """
                self._calls_mutex.lock()
                del self._calls[self._calls.index(self)]
                self._calls_mutex.unlock()

        def ui_handler ( handler, *args, **kwds ):
            """ Handles UI notification handler requests that occur on a thread other
                than the UI thread.
            """
            _CallAfter(handler, *args, **kwds)

        # Tell the traits notification handlers to use this UI handler
        set_ui_handler( ui_handler )

def qimage_to_np(qimage):
    '''
    Utility function to transorm a Qt QImage into a numpy array suitable
    for matplotlib imshow() for instance.
    '''
    import numpy as np
    from . import Qt
    w, h = qimage.width(), qimage.height()
    if isinstance(qimage, Qt.QPixmap):
        qimage = qimage.toImage()
    # sip.voidptr (qimage.bits()) asarray method is only available
    # in sip >= 4.15
    #aim = aim = np.array(qimage.bits().asarray(w * h * 4)).reshape((h, w, 4))
    b = qimage.bits()
    b.setsize(w * h * 4)
    aim = np.array(b).reshape((h, w, 4))
    # TODO: handle different pixel formats
    aim[:,:,0:3] = np.flip(aim[:,:,0:3], axis=2)
    return aim

def imshow_widget(widget, figure=None, show=False):
    '''
    Display a shapshot of a QWidget into a Matplotlib figure using
    pylab.imshow(). This is useful to use the sphinx_gallery module for
    documentation.
    '''
    from . import Qt
    from matplotlib import pyplot
    Qt.QApplication.instance().processEvents()
    if Qt.QT_VERSION >= 0x050000:
        im = widget.grab()  # Qt5 only
    else:
        im = Qt.QPixmap.grabWidget(widget)  # Qt4 only
    aim = qimage_to_np(im)
    plot = pyplot.imshow(aim, figure=figure)
    if figure is not None:
        axes = figure.axes()
    else:
        axes = pyplot.axes()
    axes.get_xaxis().set_visible(False)
    axes.get_yaxis().set_visible(False)
    if show:
        if figure is not None:
            figure.show()
        else:
            pyplot.show(block=False)
    return plot



# -*- coding: utf-8 -*-

from __future__ import print_function

# obsolete modules which seem never to be used:
#

from __future__ import absolute_import
import unittest
import os
import sys
import tempfile
import shutil
from soma import singleton
# import modules even when they are not tested, just to mark them as
# not tested in coverage tests
from soma import api
from soma import activate_virtualenv
from soma import application
from soma import archive
from soma import bufferandfile
from soma import config
from soma import controller
try:
    from soma import crypt
    have_crypt = True
except ImportError:
    # Crypto (pycrypto package) missing or outdated
    have_crypt = False
from soma import debug
from soma import factory
from soma import fom
from soma import functiontools
from soma import global_naming
from soma import html
try:
    from soma import qimage2ndarray
    from soma import qt_gui
except ImportError:
    pass # PyQt not installed
from soma import importer
from soma import info
from soma import logging
from soma import minf
from soma import notification
from soma import path
from soma import pipeline
from soma import plugins
from soma import safemkdir
from soma import sandbox
from soma import serialization
from soma import somatime
from soma import sorted_dictionary
from soma import sqlite_tools
from soma import stringtools
from soma import subprocess
from soma import test_utils
from soma import thread_calls
from soma import topological_sort
from soma import translation
from soma import undefined
from soma import utils
from soma import uuid


class TestSomaMisc(unittest.TestCase):

    def test_singleton(self):
        class ASingleton(singleton.Singleton):
            def __singleton_init__(self):
                super(ASingleton, self).__singleton_init__()
                self._shared_num = 12

        self.assertRaises(ValueError, ASingleton.get_instance)
        sing = ASingleton()
        self.assertTrue(sing is ASingleton())
        self.assertTrue(hasattr(sing, '_shared_num'))
        self.assertEqual(sing._shared_num, 12)

    if have_crypt:
        def test_crypt(self):
            private_key, public_key = crypt.generate_RSA()
            self.assertTrue(isinstance(private_key, bytes))
            self.assertTrue(isinstance(public_key, bytes))
            d = tempfile.mkdtemp()
            try:
                pubfile = os.path.join(d, 'id_rsa.pub')
                privfile = os.path.join(d, 'id_rsa')
                open(pubfile, 'wb').write(public_key)
                open(privfile, 'wb').write(private_key)

                msg = u'I write a super secret message that nobody should '\
                    'see, never.'.encode('utf-8')
                crypt_msg = crypt.encrypt_RSA(pubfile, msg)
                self.assertTrue(crypt_msg != msg)
                uncrypt_msg = crypt.decrypt_RSA(privfile, crypt_msg)
                self.assertEqual(uncrypt_msg, msg)
            finally:
                shutil.rmtree(d)

    def test_partial(self):
        def my_func(x, y, z, t, **kwargs):
            res = x + y + z + t
            if 'suffix' in kwargs:
                res += kwargs['suffix']
            return res

        p = functiontools.SomaPartial(my_func, 12, 15)
        self.assertEqual(p(10, 20), 57)
        q = functiontools.SomaPartial(my_func, 'start_', t='_t', suffix='_end')
        self.assertEqual(q('ab', z='ba'), 'start_abba_t_end')
        self.assertTrue(functiontools.hasParameter(my_func, 'y'))
        self.assertTrue(functiontools.hasParameter(my_func, 'b'))
        self.assertEqual(functiontools.numberOfParameterRange(my_func), (4, 4))

        def other_func(x, y, z, t):
            return x + y + z + t
        
        class TmpObject(object):
            def meth(self, x, y, z, t):
                return x + y + z + t

        self.assertTrue(functiontools.hasParameter(other_func, 'y'))
        self.assertFalse(functiontools.hasParameter(other_func, 'b'))
        self.assertTrue(functiontools.checkParameterCount(other_func, 4)
                        is None)
        self.assertTrue(functiontools.checkParameterCount(TmpObject().meth, 4)
                        is None)
        self.assertRaises(RuntimeError,
                          functiontools.checkParameterCount, other_func, 3)

    def test_drange(self):
        l = [x for x in functiontools.drange(2.5, 4.8, 0.6)]
        self.assertEqual(l, [2.5, 3.1, 3.7, 4.3])

    def test_archive(self):
        d = tempfile.mkdtemp()
        try:
            fullfile1 = os.path.join(d, 'archive.bop')
            open(fullfile1, 'wb').write(b'bloblop')
            self.assertFalse(archive.is_archive(fullfile1))
            dir1 = os.path.join(d, 'subdir')
            fullfile2 = os.path.join(dir1, 'archive2.txt')
            os.mkdir(dir1)
            open(fullfile2, 'w').write(u'bebert is happy')
            for ext in ('.zip', '.tar', '.tgz', 'tar.bz2'):
                arfile = os.path.join(d, 'archive' + ext)
                open(arfile, 'wb').write(b'bloblop')
                unpacked = os.path.join(d, 'unpacked')
                # the following does not behave correctly, is_archive(*.zip)
                # returns True
                #self.assertFalse(archive.is_archive(arfile))
                self.assertRaises(OSError, archive.unpack, arfile, unpacked)
                self.assertRaises(OSError, archive.unzip, arfile, unpacked)
                archive.pack(arfile, [fullfile1, dir1])
                self.assertTrue(archive.is_archive(arfile))
                try:
                    archive.unpack(arfile, unpacked)
                    self.assertTrue(os.path.isfile(
                        os.path.join(unpacked, 'archive.bop')))
                    self.assertTrue(os.path.isfile(
                        os.path.join(unpacked, 'subdir', 'archive2.txt')))
                    content1 = open(os.path.join(
                        unpacked, 'archive.bop'), 'rb').read()
                    self.assertEqual(content1, b'bloblop')
                    content2 = open(os.path.join(
                        unpacked, 'subdir', 'archive2.txt'), 'r').read()
                    self.assertEqual(content2, u'bebert is happy')
                finally:
                    try:
                        shutil.rmtree(unpacked)
                    except OSError:
                        pass
                # trunkcate archive
                content = open(arfile, 'rb').read()
                open(arfile, 'wb').write(content[:50])
                self.assertTrue(archive.is_archive(arfile)) # shoud rather fail
                unpacked = os.path.join(d, 'unpacked')
                try:
                    self.assertRaises(OSError, archive.unpack, arfile,
                                      unpacked)
                finally:
                    try:
                        shutil.rmtree(unpacked)
                    except OSError:
                        pass
                # zip one file
                archive.pack(arfile, fullfile1)
                self.assertTrue(archive.is_archive(arfile))
                try:
                    archive.unpack(arfile, unpacked)
                    self.assertTrue(os.path.isfile(
                        os.path.join(unpacked, 'archive.bop')))
                    content1 = open(os.path.join(
                        unpacked, 'archive.bop'), 'rb').read()
                    self.assertEqual(content1, b'bloblop')
                finally:
                    try:
                        shutil.rmtree(unpacked)
                    except OSError:
                        pass


        finally:
            shutil.rmtree(d)



def test():
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSomaMisc)
    runtime = unittest.TextTestRunner(verbosity=2).run(suite)
    return runtime.wasSuccessful()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Singleton pattern.

- author: Yann Cointepas
- organization: NeuroSpin
- license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import

import atexit

__docformat__ = 'restructuredtext en'


class Singleton(object):

    '''
    Implements the singleton pattern. A class deriving from ``Singleton`` can
    have only one instance. The first instanciation will create an object and
    other instanciations return the same object. Note that the :meth:`__init__`
    method (if any) is still called at each instanciation (on the same object).
    Therefore, :class:`Singleton` derived classes should define
    :meth:`__singleton_init__`
    instead of :py:meth:`__init__` because the former is only called once.

    Example::

      from singleton import Singleton

      class MyClass( Singleton ):
        def __singleton_init__( self ):
          self.attribute = 'value'

      o1 = MyClass()
      o2 = MyClass()
      print(o1 is o2)

    '''

    @classmethod
    def get_instance(cls):
        try:
            return getattr(cls, '_singleton_instance')
        except AttributeError:
            msg = "Class %s has not been initialized" % cls.__name__
            raise ValueError(msg)

    def __new__(cls, *args, **kwargs):
        if '_singleton_instance' not in cls.__dict__:
            cls._singleton_instance = super(Singleton, cls).__new__(cls)
            singleton_init = getattr(cls._singleton_instance,
                                     '__singleton_init__', None)
            if singleton_init is not None:
                singleton_init(*args, **kwargs)
            atexit.register(cls.delete_singleton)
        return cls._singleton_instance

    def __init__(self, *args, **kwargs):
        '''
        The __init__ method of :py:class:`Singleton` derived class should do
        nothing.
        Derived classes must define :py:meth:`__singleton_init__` instead of __init__.
        '''

    def __singleton_init__(self, *args, **kwargs):
        super(Singleton, self).__init__(*args, **kwargs)

    @classmethod
    def delete_singleton(cls):
        if hasattr(cls, '_singleton_instance'):
            del cls._singleton_instance



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

# System import
from __future__ import absolute_import
import os
import logging
from functools import partial
from traits.api import Instance
import six

# Define the logger
logger = logging.getLogger(__name__)

# Soma import
from soma.qt_gui.qt_backend import QtGui, QtCore
from soma.utils.functiontools import SomaPartial
from soma.controller import trait_ids
from soma.controller import Controller
from soma.sorted_dictionary import OrderedDict
from soma.qt_gui.controller_widget import ControllerWidget, weak_proxy
import traits.api as traits
import sip

# Qt import
try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

QtCore.QResource.registerResource(os.path.join(os.path.dirname(
    os.path.dirname(__file__)), 'resources', 'widgets_icons.rcc'))


class DictController(Controller):

    """ Dummy dict controller to simplify the creation of a dict widget
    """
    pass


class DictControlWidget(object):

    """ Control to enter a dict of items.
    """

    #
    # Public members
    #

    @staticmethod
    def is_valid(control_instance, *args, **kwargs):
        """ Method to check if the new control values are correct.

        If the new dict controls values are not correct, the backroung
        color of each control in the dict will be red.

        Parameters
        ----------
        control_instance: QFrame (mandatory)
            the control widget we want to validate

        Returns
        -------
        valid: bool
            True if the control values are valid,
            False otherwise
        """
        # Initilaized the output
        valid = True

        # Go through all the controller widget controls
        controller_widget = control_instance.controller_widget
        for control_name, control_groups \
                in six.iteritems(controller_widget._controls):

            if not control_groups:
                continue
            # Unpack the control item
            trait, control_class, control_instance, control_label \
                = next(iter(control_groups.values()))

            # Call the current control specific check method
            valid = control_class.is_valid(control_instance)

            # Stop checking if a wrong control has been found
            if not valid:
                break

        return valid

    @classmethod
    def check(cls, control_instance):
        """ Check if a controller widget dict control is filled correctly.

        Parameters
        ----------
        cls: DictControlWidget (mandatory)
            a DictControlWidget control
        control_instance: QFrame (mandatory)
            the control widget we want to validate
        """
        pass

    @staticmethod
    def add_callback(callback, control_instance):
        """ Method to add a callback to the control instance when the dict
        trait is modified

        Parameters
        ----------
        callback: @function (mandatory)
            the function that will be called when a 'textChanged' signal is
            emited.
        control_instance: QFrame (mandatory)
            the control widget we want to validate
        """
        pass

    @staticmethod
    def create_widget(parent, control_name, control_value, trait,
                      label_class=None, user_data=None):
        """ Method to create the dict widget.

        Parameters
        ----------
        parent: QWidget (mandatory)
            the parent widget
        control_name: str (mandatory)
            the name of the control we want to create
        control_value: dict of items (mandatory)
            the default control value
        trait: Tait (mandatory)
            the trait associated to the control
        label_class: Qt widget class (optional, default: None)
            the label widget will be an instance of this class. Its constructor
            will be called using 2 arguments: the label string and the parent
            widget.

        Returns
        -------
        out: 2-uplet
            a two element tuple of the form (control widget: ,
            associated labels: (a label QLabel, the tools QWidget))
        """
        # Get the inner trait: expect only one inner trait
        # note: trait.inner_traits might be a method (ListInt) or a tuple
        # (List), whereas trait.handler.inner_trait is always a method
        if len(trait.handler.inner_traits()) != 2:
            raise Exception(
                "Expect two inner traits in Dict control. Trait '{0}' "
                "inner traits are '{1}'.".format(
                    control_name, trait.inner_traits))
        inner_trait = trait.handler.inner_traits()[1]

        # Create the dict widget: a frame
        frame = QtGui.QFrame(parent=parent)
        frame.setFrameShape(QtGui.QFrame.StyledPanel)
        frame.user_data = user_data

        # Create tools to interact with the dict widget: expand or collapse -
        # add a dict item - remove a dict item
        tool_widget = QtGui.QWidget(parent)
        layout = QtGui.QHBoxLayout()
        layout.addStretch(1)
        tool_widget.setLayout(layout)
        # Create the tool buttons
        resize_button = QtGui.QToolButton()
        add_button = QtGui.QToolButton()
        layout.addWidget(resize_button)
        layout.addWidget(add_button)
        # Set the tool icons
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(_fromUtf8(":/soma_widgets_icons/add")),
                       QtGui.QIcon.Normal, QtGui.QIcon.Off)
        add_button.setIcon(icon)
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(_fromUtf8(":/soma_widgets_icons/nav_down")),
            QtGui.QIcon.Normal, QtGui.QIcon.Off)
        resize_button.setIcon(icon)
        resize_button.setFixedSize(30, 22)
        add_button.setFixedSize(30, 22)

        # Create a new controller that contains length 'control_value' inner
        # trait elements
        controller = DictController()
        for name, inner_control_values in six.iteritems(control_value):
            controller.add_trait(str(name), inner_trait)
            setattr(controller, str(name), inner_control_values)

        # Create the associated controller widget
        controller_widget = ControllerWidget(controller, parent=frame,
                                             live=True, editable_labels=True,
                                             user_data=user_data)

        # Store some parameters in the dict widget
        frame.inner_trait = inner_trait
        frame.trait = trait
        frame.controller = controller
        frame.controller_widget = controller_widget
        frame.connected = False

        # Add the dict controller widget to the dict widget
        frame.setLayout(controller_widget.layout())

        # Set some callback on the dict control tools
        # Resize callback
        resize_hook = partial(
            DictControlWidget.expand_or_collapse, weak_proxy(frame),
            weak_proxy(resize_button))
        resize_button.clicked.connect(resize_hook)
        # Add dict item callback
        add_hook = partial(
            DictControlWidget.add_dict_item, parent, control_name, frame)
        add_button.clicked.connect(add_hook)

        # Create the label associated with the dict widget
        control_label = trait.label
        if control_label is None:
            control_label = control_name
        if label_class is None:
            label_class = QtGui.QLabel
        if control_label is not None:
            label = label_class(control_label, parent)
        else:
            label = None

        controller_widget.main_controller_def = (DictControlWidget, parent,
                                                 control_name, frame)
        return (frame, (label, tool_widget))

    @staticmethod
    def update_controller(controller_widget, control_name, control_instance,
                          *args, **kwarg):
        """ Update one element of the controller.

        At the end the controller trait value with the name 'control_name'
        will match the controller widget user parameters defined in
        'control_instance'.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Get the dict widget inner controller values
        new_trait_value = dict([
            (name, getattr(control_instance.controller, name))
            for name in control_instance.controller.user_traits()])

        protected = controller_widget.controller.is_parameter_protected(
            control_name)
        # value is manually modified: protect it
        if getattr(controller_widget.controller, control_name) \
                != new_trait_value:
            controller_widget.controller.protect_parameter(control_name)
        # Update the 'control_name' parent controller value
        try:
            setattr(controller_widget.controller, control_name,
                    new_trait_value)
        except (traits.TraitError, TypeError):
            print('invalid value set in dict %s:' % control_name,
                  new_trait_value)
            if not protected:
                controller_widget.controller.unprotect_parameter(control_name)
        logger.debug(
            "'DictControlWidget' associated controller trait '{0}' has "
            "been updated with value '{1}'.".format(
                control_name, new_trait_value))

    @classmethod
    def update_controller_widget(cls, controller_widget, control_name,
                                 control_instance):
        """ Update one element of the dict controller widget.

        At the end the dict controller widget user editable parameter with the
        name 'control_name' will match the controller trait value with the same
        name.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        try:
            was_connected = control_instance.connected
        except ReferenceError:
            # widget deleted in the meantime
            return

        if sip.isdeleted(control_instance.__init__.__self__):
            DictControlWidget.disconnect(controller_widget, control_name,
                                         control_instance)
            return

        # One callback has not been removed properly
        if control_name in controller_widget.controller.user_traits():

            # Get the dict widget current connection status
            was_connected = control_instance.connected

            # Disconnect the dict controller and the inner dict controller
            cls.disconnect(controller_widget, control_name, control_instance)
            control_instance.controller_widget.disconnect()

            # Get the 'control_name' dict value from the top dict controller
            trait_value = getattr(controller_widget.controller, control_name)

            # Get the number of dict elements in the controller associated
            # with the current dict control
            widget_traits = control_instance.controller.user_traits()

            # Parameter that is True if a user trait associated with the
            # current dict control has changed
            user_traits_changed = False

            # Special case: some traits have been deleted to the top controller
            # Need to remove to the inner dict controller some traits
            for trait_name in widget_traits:
                if trait_name not in trait_value:
                    control_instance.controller.remove_trait(trait_name)
                    # Notify that some traits of the inner dict controller have
                    # been deleted
                    user_traits_changed = True

            # Special case: some new traits have been added to the top
            # controller
            # Need to add to the inner dict controller some traits
            # with type 'inner_trait'
            for trait_name in trait_value:
                if trait_name not in widget_traits:
                    control_instance.controller.add_trait(
                        trait_name, control_instance.inner_trait)
                    # Notify that some traits of the inner dict controller
                    # have been added
                    user_traits_changed = True

            # Update the controller associated with the current control
            for trait_name, value in six.iteritems(trait_value):
                setattr(control_instance.controller, trait_name, value)

            # Connect the inner dict controller
            control_instance.controller_widget.connect()

            # Emit the 'user_traits_changed' signal if necessary
            if user_traits_changed:
                control_instance.controller.user_traits_changed = True

                logger.debug(
                    "'DictControlWidget' inner controller has been updated:"
                    "old size '{0}', new size '{1}'.".format(
                        len(widget_traits), len(trait_value)))

            # Restore the previous dict controller connection status
            if was_connected:
                cls.connect(controller_widget, control_name, control_instance)

        else:
            logger.error("oups")
            # print cls, controller_widget, control_name, control_instance
            # print control_instance.controller
            # print control_instance.controller.user_traits()

    @classmethod
    def connect(cls, controller_widget, control_name, control_instance):
        """ Connect a 'List' controller trait and a 'DictControlWidget'
        controller widget control.

        Parameters
        ----------
        cls: StrControlWidget (mandatory)
            a StrControlWidget control
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str (mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Check if the control is connected
        if not control_instance.connected:

            # Update the dict item when one of his associated controller trait
            # changed.
            # Hook: function that will be called to update the controller
            # associated with a dict widget when a dict widget inner controller
            # trait is modified.
            dict_controller_hook = SomaPartial(
                cls.update_controller, weak_proxy(controller_widget),
                control_name, weak_proxy(control_instance))

            # Go through all dict widget inner controller user traits
            for trait_name in control_instance.controller.user_traits():

                # And add the callback on each user trait
                control_instance.controller.on_trait_change(
                    dict_controller_hook, trait_name)
                logger.debug("Item '{0}' of a 'DictControlWidget', add "
                             "a callback on inner controller trait "
                             "'{0}'.".format(control_name, trait_name))

            # Update the dict controller widget.
            # Hook: function that will be called to update the specific widget
            # when a trait event is detected on the dict controller.
            controller_hook = SomaPartial(
                cls.update_controller_widget, weak_proxy(controller_widget),
                control_name, weak_proxy(control_instance))

            # When the 'control_name' controller trait value is modified,
            # update the corresponding control
            controller_widget.controller.on_trait_change(
                controller_hook, control_name, dispatch='ui')

            # Update the dict connection status
            control_instance._controller_connections = (
                dict_controller_hook, controller_hook)
            logger.debug("Add 'Dict' connection: {0}.".format(
                control_instance._controller_connections))

            # Connect also all dict items
            inner_controls = control_instance.controller_widget._controls
            for (inner_control_name,
                 inner_control_groups) in six.iteritems(inner_controls):
                for group, inner_control \
                        in six.iteritems(inner_control_groups):

                    # Unpack the control item
                    inner_control_instance = inner_control[2]
                    inner_control_class = inner_control[1]

                    # Call the inner control connect method
                    inner_control_class.connect(
                        control_instance.controller_widget, inner_control_name,
                        inner_control_instance)

            # Update the dict control connection status
            control_instance.connected = True

    @staticmethod
    def disconnect(controller_widget, control_name, control_instance):
        """ Disconnect a 'List' controller trait and a 'DictControlWidget'
        controller widget control.

        Parameters
        ----------
        cls: StrControlWidget (mandatory)
            a StrControlWidget control
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str (mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Check if the control is connected
        if control_instance.connected:

            # Get the stored widget and controller hooks
            (dict_controller_hook,
             controller_hook) = control_instance._controller_connections

            # Remove the controller hook from the 'control_name' trait
            controller_widget.controller.on_trait_change(
                controller_hook, control_name, remove=True)

            # Remove the dict controller hook associated with the controller
            # traits
            for trait_name in control_instance.controller.user_traits():
                control_instance.controller.on_trait_change(
                    dict_controller_hook, trait_name, remove=True)

            # Delete the trait - control connection we just remove
            del control_instance._controller_connections

            # Disconnect also all dict items
            inner_controls = control_instance.controller_widget._controls
            for (inner_control_name,
                 inner_control_groups) in six.iteritems(inner_controls):
                for group, inner_control \
                        in six.iteritems(inner_control_groups):

                    # Unpack the control item
                    inner_control_instance = inner_control[2]
                    inner_control_class = inner_control[1]

                    # Call the inner control disconnect method
                    inner_control_class.disconnect(
                        control_instance.controller_widget, inner_control_name,
                        inner_control_instance)

            # Update the dict control connection status
            control_instance.connected = False

    #
    # Callbacks
    #

    @staticmethod
    def add_dict_item(controller_widget, control_name, control_instance):
        """ Append one element in the dict controller widget.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Get a new key name
        trait_name = 'new_item'
        i = 1
        while control_instance.controller.trait(trait_name):
            trait_name = 'new_item_%d' % i
            i += 1

        was_connected = controller_widget.connected
        if was_connected:
            controller_widget.disconnect()
        # Add the new trait to the inner list controller
        control_instance.controller.add_trait(
            trait_name, control_instance.inner_trait)
        DictControlWidget.update_controller(
            controller_widget, control_name, control_instance)
        if was_connected:
            controller_widget.connect()

        # update interface
        control_instance.controller_widget.update_controls()

        logger.debug("Add 'DictControlWidget' '{0}' new trait "
                     "callback.".format(trait_name))

    @staticmethod
    def expand_or_collapse(control_instance, resize_button):
        """ Callback to expand or collapse a 'DictControlWidget'.

        Parameters
        ----------
        control_instance: QFrame (mandatory)
            the dict widget item
        resize_button: QToolButton
            the signal sender
        """
        # Change the icon depending on the button status
        icon = QtGui.QIcon()

        # Hide the control
        if control_instance.isVisible():
            control_instance.hide()
            icon.addPixmap(QtGui.QPixmap(_fromUtf8(
                ":/soma_widgets_icons/nav_right")),
                QtGui.QIcon.Normal, QtGui.QIcon.Off)

        # Show the control
        else:
            control_instance.show()
            icon.addPixmap(QtGui.QPixmap(_fromUtf8(
                ":/soma_widgets_icons/nav_down")),
                QtGui.QIcon.Normal, QtGui.QIcon.Off)

        # Set the new button icon
        resize_button.setIcon(icon)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

from soma.qt_gui.qt_backend import Qt


class QVTabBar(Qt.QTabBar):
    '''
    A vertical QTabBar, as in
    https://stackoverflow.com/questions/50578661/how-to-implement-vertical-tabs-in-qt
    '''

    def __init__(self, parent=None, tab_width=None, tab_height=None):
        super(QVTabBar, self).__init__(parent)
        self.tab_width = tab_width
        self.tab_height = tab_height
        self._first_paint = True

    def tabSizeHint(self, index):
        s = Qt.QTabBar.tabSizeHint(self, index)
        s.transpose()
        if self.tab_width is not None:
            s.setWidth(self.tab_width)
        if self.tab_height is not None:
            s.setHeight(self.tab_height)
        return s

    def paintEvent(self, event):
        if self._first_paint:
            self._first_paint = True
            self.update_buttons()

        painter = Qt.QStylePainter(self)
        opt = Qt.QStyleOptionTab()

        for i in range(self.count()):
            self.initStyleOption(opt, i)
            painter.drawControl(Qt.QStyle.CE_TabBarTabShape, opt)
            painter.save()

            s = opt.rect.size()
            s.transpose()
            r = Qt.QRect(Qt.QPoint(), s)
            r.moveCenter(opt.rect.center())
            opt.rect = r

            c = self.tabRect(i).center()
            painter.translate(c)
            painter.rotate(90)
            painter.translate(-c)
            painter.drawControl(Qt.QStyle.CE_TabBarTabLabel, opt)
            painter.restore()


        Qt.QWidget.paintEvent(self, event)

    def resize_optimal(self):
        sizes = [self.tabSizeHint(i) for i in range(self.count())]
        width = max([s.width() for s in sizes])
        height = sum([s.height() for s in sizes])
        self.resize(width, height)

    def update_buttons(self):
        # check if tab is closable
        if self.tabsClosable():
            opt = Qt.QStyleOptionTab()

            for i in range(self.count()):
                self.initStyleOption(opt, i)
                optRect = opt.rect;
                optRect.setX(optRect.width() - 14)  # set X pos of close button
                optRect.setY(optRect.y() + 8)  # calcs the Y pos of close button
                optRect.setSize(Qt.QSize(12, 12))
                self.tabButton(i, Qt.QTabBar.RightSide).setGeometry(optRect)



class QVTabWidget(Qt.QTabWidget):
    '''
    A vertical QTabWidget, as in
    https://stackoverflow.com/questions/50578661/how-to-implement-vertical-tabs-in-qt
    '''

    def __init__(self, parent=None, width=None, height=None):
        super(QVTabWidget, self).__init__(parent)
        self.setTabBar(QVTabBar(parent, width, height))
        self.setTabPosition(Qt.QTabWidget.West)

    def setTabControlSize(self, width, height):
        self.tabBar().tab_width = width
        self.tabBar().tab_height = height
        self.tabBar().resize_optimal()
        self.resize(self.sizeHint())

    def update_buttons(self):
        self.tabBar().update_buttons()



# -*- coding: utf-8 -*-

from __future__ import print_function

from __future__ import absolute_import
import unittest
from soma import html



class TestHtml(unittest.TestCase):

    def test_html_escape(self):
        s1 = u'totô < 0 ét tùtu bàbär >= 3 & etc.'
        s2 = html.htmlEscape(s1)
        self.assertEqual(
            s2,
            u'tot&ocirc; &lt; 0 &eacute;t t&ugrave;tu b&agrave;b&auml;r &gt;'
            '= 3 &amp; etc.')
        s3 = html.lesserHtmlEscape(s1)
        self.assertEqual(
            s3,
            u'totô &lt; 0 ét tùtu bàb&auml;r &gt;= 3 &amp; etc.')


def test():
    suite = unittest.TestLoader().loadTestsFromTestCase(TestHtml)
    runtime = unittest.TextTestRunner(verbosity=2).run(suite)
    return runtime.wasSuccessful()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-

""" Config module that sets up version variable and finds the
`BrainVISA <http://brainvisa.info>`_ `brainvisa-share` data directory.

Attributes
----------
fullVersion
shortVersion
full_version
short_version
BRAINVISA_SHARE
"""

from __future__ import absolute_import
import os
import sys
import soma.info

full_version = ".".join([str(soma.info.version_major),
                        str(soma.info.version_minor),
                        str(soma.info.version_micro)])
short_version = ".".join([str(soma.info.version_major),
                         str(soma.info.version_minor)])

fullVersion = full_version
shortVersion = short_version


def _init_default_brainvisa_share():
    try:
        import brainvisa_share.config
        bv_share_dir = brainvisa_share.config.share
        has_config = True
    except ImportError:
        bv_share_dir = "brainvisa-share-%s" % short_version
        has_config = False

    if bv_share_dir and os.path.exists(bv_share_dir):
        return bv_share_dir

    share = os.getenv('BRAINVISA_SHARE')
    if share:
        # share is the base share/ directory: we must find the brainvisa-share
        # directory in it.
        share = os.path.join(share, bv_share_dir)
    if not share or not os.path.exists(share):
        if has_config:
            share = os.path.join(os.path.dirname(os.path.dirname(
                os.path.dirname(
                    brainvisa_share.config.__file__))), 'share',
                    brainvisa_share.config.share)
        else:
            share = os.path.join(
                os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
                'share', bv_share_dir)  # cannot do better.
    return share

BRAINVISA_SHARE = _init_default_brainvisa_share()



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
* author: Nicolas Souedet
* organization: NeuroSpin
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_

Import subprocess32 or subprocess API depending on python version and what is
available on the system.
'''

from __future__ import absolute_import
import sys

if sys.version_info[:2] >= (3, 2):
    # in python >= 3.2, subprocess32 is not needed as it is the builtin
    # subprocess module
    from subprocess import *
else:
    try:
        def __initialize_zmq():
            # It is necessary to first import zmq from the system if it is
            # installed otherwise the one embeded with subprocess32 is loaded
            # and it can lead to compatibility issue
            import zmq

        __initialize_zmq()

        del __initialize_zmq

    except ImportError:
        pass

    try:
        # It is necessary to replace subprocess by subprocess32 to fix issue
        # in subprocess start
        # Import in current module all that is defined in subprocess module
        from subprocess32 import *

        def __initialize_subprocess32():
            import subprocess32
            import subprocess as _subprocess
            if hasattr(_subprocess, '_args_from_interpreter_flags'):
                # get this private function which is used somewhere in
                # multiprocessing
                subprocess32._args_from_interpreter_flags \
                    = _subprocess._args_from_interpreter_flags
            del _subprocess
            import sys
            sys.modules['subprocess'] = sys.modules['subprocess32']

        __initialize_subprocess32()
        del __initialize_subprocess32

    except ImportError:
        from subprocess import *

        def __initialize_subprocess():
            import subprocess

        __initialize_subprocess()
        del __initialize_subprocess



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
See :py:mod:`minf.api <soma.minf.api>` module for documentation.

* author: Yann Cointepas
* organization: `NeuroSpin <http://www.neurospin.org>`_
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
__docformat__ = "restructuredtext en"



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import tarfile
import zipfile
import os

def is_archive(filename):
    """
    Returns true if filename is an archive and false otherwise.
    """
    if zipfile.is_zipfile(filename):
        return True
    else:
        try:
            tarfile.open(filename)
            return True
        except Exception:
            pass
    if os.path.splitext(filename)[1] in get_archive_extensions():
        return True
    return False

def get_archive_extensions():
    return ['.zip', '.gz', '.tar', '.bz2', '.tgz']

def unpack(input_filename, extract_dir):
    """
    Unpacks the input_filename archive to the extract_dir directory.
    """
    if zipfile.is_zipfile(input_filename):
        unzip(input_filename, extract_dir)
    else:
        untar(input_filename, extract_dir)

def pack(output_filename, sources):
    """
    Packs the source_dir directory in the output_filename archive.
    """
    ext = os.path.splitext(output_filename)[1][1:]
    if ext == 'zip':
        pack_zip(output_filename, sources)
    elif ext == 'gz' or ext == 'tgz' or ext == 'bz2' or ext == 'tar':
        pack_tar(output_filename, sources, ext)
    
def untar(input_filename, extract_dir):
    """
    Extracts the input_filename archive to the extract_dir directory.
    """
    try:
        tar_ds = tarfile.open(input_filename)
    except (tarfile.TarError, EOFError):
        raise OSError("%s is not a tar file" % (input_filename))
    tar_ds.extractall(path=extract_dir)
    tar_ds.close()

def unzip(input_filename, extract_dir):
    """
    Extracts the input_filename archive to the extract_dir directory.
    """
    if not zipfile.is_zipfile(input_filename):
        raise OSError("%s is not a zip file" % (input_filename))
    zip_ds = zipfile.ZipFile(input_filename)
    zip_ds.extractall(path=extract_dir)
    zip_ds.close()

def pack_tar(output_filename, sources, type='gz'):
    """
    Creates a tar archive in output_filename from the source_dir directory.
    """
    if type == 'tgz':
        type = 'gz'
    elif type == 'tar':
        type = ''
    tar_ds = tarfile.open(output_filename, 'w:' + type)
    if not isinstance(sources, (list, tuple)) and \
       isinstance(sources, str):
        sources = [sources]
    for source in sources:
        tar_ds.add(source, arcname=os.path.basename(source))
    tar_ds.close()

def pack_zip(output_filename, sources):
    """
    Creates a zip archive in output_filename from the source_dir directory.
    """
    previous_dir = os.getcwd()
    if not isinstance(sources, (list, tuple)) and \
       isinstance(sources, str):
        sources = [sources]
    zip_ds = zipfile.ZipFile(output_filename, 'w', zipfile.ZIP_DEFLATED)
    for source in sources:
        os.chdir(os.path.dirname(source))
        if os.path.isdir(source):
            for root, dirs, files in os.walk(os.path.basename(source)):
                for file in files:
                    zip_ds.write(os.path.join(root, file))
        else:
            zip_ds.write(os.path.basename(source))
    zip_ds.close()
    os.chdir(previous_dir)



# -*- coding: utf-8 -*-
from .base_classes import test_modes, run_mode, ref_mode, default_mode, \
    SomaTestLoader, SomaTestCase

__all__ = [test_modes, run_mode, ref_mode, default_mode, SomaTestLoader,
           SomaTestCase]



# -*- coding: utf-8 -*-

from __future__ import print_function

from __future__ import absolute_import
import unittest
import shutil
import os
import tempfile
from soma.controller import Controller, ControllerTrait, OpenKeyController
import traits.api as traits
from soma.controller.trait_utils import (
    get_trait_desc, is_trait_value_defined, is_trait_pathname,
    trait_ids)
from soma.controller import factory



class TestController(unittest.TestCase):

    def test_controller(self):
        c1 = Controller()
        c1.add_trait('gogo', traits.Str())
        c1.add_trait('bozo', traits.Int(12))
        self.assertEqual(c1.gogo, '')
        self.assertEqual(c1.bozo, 12)
        self.assertEqual(list(c1.user_traits().keys()), ['gogo', 'bozo'])
        c1.gogo = 'blop krok'
        self.assertEqual(c1.gogo, 'blop krok')
        d = c1.export_to_dict()
        self.assertEqual(d, {'gogo': 'blop krok', 'bozo': 12})
        c1.reorder_traits(['bozo', 'gogo'])
        self.assertEqual(list(c1.user_traits().keys()), ['bozo', 'gogo'])
        c1.reorder_traits(['gogo', 'bozo'])
        self.assertEqual(list(c1.user_traits().keys()), ['gogo', 'bozo'])

    def test_controller2(self):
        class Zuzur(Controller):
            glop = traits.Str('zut')

        c2 = Zuzur()
        c3 = Zuzur()
        self.assertEqual(c2.glop, 'zut')
        c2.glop = 'I am c2'
        c3.glop = 'I am c3'
        self.assertEqual(c2.glop, 'I am c2')
        self.assertEqual(c3.glop, 'I am c3')

    def test_controller3(self):
        class Babar(Controller):
            hupdahup = traits.Str('barbatruc')
            gargamel = traits.Str(traits.Undefined)
            ouioui = traits.List(traits.Str())

        c1 = Babar()
        self.assertEqual(c1.gargamel, traits.Undefined)
        d = c1.export_to_dict()
        self.assertEqual(d, {'hupdahup': 'barbatruc',
                             'gargamel': traits.Undefined,
                             'ouioui': []})
        c2 = Babar()
        c2.gargamel = 'schtroumpf'
        c2.import_from_dict(d)
        self.assertEqual(c2.export_to_dict(), d)
        d = c1.export_to_dict(exclude_undefined=True)
        self.assertEqual(d, {'hupdahup': 'barbatruc', 'ouioui': []})
        c2.gargamel = 'schtroumpf'
        c2.import_from_dict(d)
        self.assertEqual(c2.export_to_dict(exclude_empty=True),
                         {'hupdahup': 'barbatruc', 'gargamel': 'schtroumpf'})

    def test_controller4(self):
        class Driver(Controller):
            head = traits.Str()
            arms = traits.Str()
            legs = traits.Str()

        class Car(Controller):
            wheels = traits.Str()
            engine = traits.Str()
            driver = ControllerTrait(Driver(),
                                     desc='the guy who would better take a '
                                     'bus')
            problems = ControllerTrait(OpenKeyController(traits.Str()))

        my_car = Car()
        my_car.wheels = 'flat'
        my_car.engine = 'wind-broken'
        my_car.driver.head = 'empty' # ! modify class trait !
        my_car.driver.arms = 'heavy'
        my_car.driver.legs = 'short'
        my_car.problems = {'exhaust': 'smoking', 'windshield': 'cracked'}

        d = my_car.export_to_dict()
        self.assertEqual(d, {'wheels': 'flat', 'engine': 'wind-broken',
                             'driver': {'head': 'empty', 'arms': 'heavy',
                                        'legs': 'short'},
                             'problems': {'exhaust': 'smoking',
                                          'windshield': 'cracked'}})
        self.assertTrue(isinstance(my_car.driver, Driver))
        self.assertTrue(isinstance(my_car.problems, OpenKeyController))
        my_car.driver = {'head': 'smiling', 'legs': 'strong'}
        d = my_car.export_to_dict()
        self.assertEqual(d, {'wheels': 'flat', 'engine': 'wind-broken',
                             'driver': {'head': 'smiling', 'arms': '',
                                        'legs': 'strong'},
                             'problems': {'exhaust': 'smoking',
                                          'windshield': 'cracked'}})

        other_car = my_car.copy(with_values=True)
        self.assertEqual(other_car.export_to_dict(), d)
        other_car = my_car.copy(with_values=False)
        self.assertEqual(other_car.export_to_dict(),
                         {'wheels': '', 'engine': '',
                          'driver': {'head': 'empty', 'arms': 'heavy',
                                     'legs': 'short'},
                          'problems': {}})

        self.assertRaises(traits.TraitError,
                          setattr, my_car.problems, 'fuel', 3.5)
        del my_car.problems.fuel
        self.assertEqual(sorted(my_car.problems.user_traits().keys()),
                         ['exhaust', 'windshield'])

        manhelp = get_trait_desc('driver', my_car.trait('driver'))
        self.assertEqual(
            manhelp[0],
            "driver: a legal value (['ControllerTrait'] - mandatory)")
        self.assertEqual(manhelp[1], "    the guy who would better take a bus")


    def test_trait_utils1(self):
        """ Method to test if we can build a string description for a trait.
        """
        trait = traits.CTrait(0)
        trait.handler = traits.Float()
        trait.ouptut = False
        trait.optional = True
        trait.desc = "bla"
        manhelp = get_trait_desc("float_trait", trait, 5)
        self.assertEqual(
            manhelp[0],
            "float_trait: a float (['Float'] - optional, default value: 5)")
        self.assertEqual(manhelp[1], "    bla")

    def test_trait_utils2(self):
        trait = traits.CTrait(0)
        trait.handler = traits.Float()
        trait.ouptut = True
        trait.optional = False
        manhelp = get_trait_desc("float_trait", trait, 5)
        self.assertEqual(
            manhelp[0],
            "float_trait: a float (['Float'] - mandatory, default value: 5)")
        self.assertEqual(manhelp[1], "    No description.")

    def test_trait_utils3(self):
        class Blop(object):
            pass
        trait = traits.CTrait(0)
        trait.handler = traits.Instance(Blop())
        trait.ouptut = False
        trait.optional = False
        manhelp = get_trait_desc("blop", trait, None)
        desc = ' '.join([x.strip() for x in manhelp[:-1]])
        self.assertEqual(
            desc,
            "blop: a Blop or None (['Instance_%s.Blop'] - mandatory)"
            % Blop.__module__)
        self.assertEqual(manhelp[-1], "    No description.")

    def test_trait_utils4(self):
        trait = traits.Either(traits.Int(47), traits.Str("vovo")).as_ctrait()
        trait.ouptut = False
        trait.optional = False
        manhelp = get_trait_desc("choice", trait, None)
        desc = ' '.join([x.strip() for x in manhelp[:-1]])
        self.assertTrue(
            desc in ("choice: an integer (int or long) or a string "
                     "(['Int', 'Str'] - mandatory)",
                     "choice: an integer or a string "
                     "(['Int', 'Str'] - mandatory)"))
        self.assertEqual(manhelp[-1], "    No description.")


    def test_trait(self):
        """ Method to test trait characterisitics: value, type.
        """
        self.assertTrue(is_trait_value_defined(5))
        self.assertFalse(is_trait_value_defined(""))
        self.assertFalse(is_trait_value_defined(None))
        self.assertFalse(is_trait_value_defined(traits.Undefined))

        trait = traits.CTrait(0)
        trait.handler = traits.Float()
        self.assertFalse(is_trait_pathname(trait))
        for handler in [traits.File(), traits.Directory()]:
            trait.handler = handler
            self.assertTrue(is_trait_pathname(trait))




def test():
    suite = unittest.TestLoader().loadTestsFromTestCase(TestController)
    runtime = unittest.TextTestRunner(verbosity=2).run(suite)
    return runtime.wasSuccessful()


if __name__ == "__main__":
    test()
